package org.processmining.plugins.resourcePattern;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;

import javax.swing.JRadioButton;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.resourcePattern.presentation.ConfigurationPanelWorker;
import org.processmining.plugins.resourcePattern.presentation.Viewer;
import org.processmining.plugins.resourcePattern.queueAnalysis.presentation.mvc.QueueController;
import org.processmining.plugins.resourcePattern.queueAnalysis.presentation.mvc.QueueViewer;
import org.processmining.plugins.resourcePattern.utility.Event;
import org.processmining.plugins.resourcePattern.utility.Helper;
import org.processmining.plugins.resourcePattern.utility.Parameter;


@Plugin(
	name = "Resource Queue Behaviour", 
	parameterLabels = { "Event Log"}, 
	returnLabels = { "Resource Queue Behaviour Analysis" }, 
	returnTypes = { Viewer.class }, 
	userAccessible = true, 
	help = "The plug-in takes existing XES log and detect the queue discipline employed by resources."
			
)
public class ResourcePattern {

	@PluginVariant(variantLabel = "Resource Queue Behaviour", 
			requiredParameterLabels = {0})
	@UITopiaVariant(affiliation = "Queensland University of Technology", 
			author = "Suriadi", email = "s.suriadi@qut.edu.au")
	public Viewer defaultEventGapAnalysis(UIPluginContext context,XLog log) throws Exception {
		
		//Obtain the necessary display to retrieve configuration parameters
		ConfigurationPanelWorker configWorker = new ConfigurationPanelWorker();
		Helper helper = new Helper();
		
		String[] parameterNames = new String[3];
		String[] parameterDesc = new String[3];
		parameterNames[0] = ConstantValues.paramName01startDate;
		parameterNames[1] = ConstantValues.paramName02endDate;
		//parameterNames[2] = ConstantValues.paramName03HeadType;
		//parameterNames[3] = ConstantValues.paramName03ATailType;
		parameterNames[2] = ConstantValues.paramName04ResourceAnalysisType;
		//parameterNames[5] = ConstantValues.paramName_QQueueType;
		//parameterNames[6] = ConstantValues.paramName_QScopeType;
		//parameterNames[7] = ConstantValues.paramName_QPriorityAttribute;
		parameterDesc[0] = ConstantValues.paramDesc01startDate;
		parameterDesc[1] = ConstantValues.paramDesc02endDate;
		//parameterDesc[2] = ConstantValues.paramDesc03HeadType;
		//parameterDesc[3] = ConstantValues.paramDesc03ATailType;
		parameterDesc[2] = ConstantValues.paramDesc04ResourceAnalysisType;
		//parameterDesc[5] = ConstantValues.paramDesc_QQueueType;
		//parameterDesc[6] = ConstantValues.paramDesc_QScopeType;
		//parameterDesc[7] = ConstantValues.paramDesc_QPriorityAttribute;
		HashMap<String, Object> paramSet = configWorker.prepareConfiguration(context, 
				parameterNames, parameterDesc, log);
		
		//Retrieve and clean the parameter input
		Parameter parameter = new Parameter();
		HashMap<String, Object> cleanParam = parameter.processParameter(paramSet);
		Date startDate = new Date ( ((Date) cleanParam.get(parameterNames[0])).getTime() - (24*60*60*1000) );
		Date endDate = new Date (   ((Date) cleanParam.get(parameterNames[1])).getTime() + (24*60*60*1000) );
		//String beginTransactionType = (String) cleanParam.get(parameterNames[2]);
		//String endTransactionType = (String) cleanParam.get(parameterNames[3]);
		JRadioButton resourceAnalysisType =  (JRadioButton) cleanParam.get(parameterNames[2]);
		//String queueType = (String) cleanParam.get(parameterNames[5]);
		//String scopeType = (String) cleanParam.get(parameterNames[6]);
		//String attributeName = (String) cleanParam.get(parameterNames[7]);
		
		//Obtain all events within the time range provided by the parameter
		context.log("Sorting the event log ... (Stage 1 out of 3)");
		HashSet<String> eventAttributeNames = helper.populateEventAttributeNames(log);
		ArrayList<Event> allWithinEvents = helper.getEventsPerTimePeriod(log, 
				startDate, endDate, eventAttributeNames);	
					
		context.getProgress().setValue(0);
		context.getProgress().setMaximum(300);
		
		context.log("Calculating gaps between events... (Stage 2 out of 3");
		String gapPath = ConstantValues.filePath;
		/*Gap gap = new Gap();
		//System.out.println("here here here 7");
		if (resourceAnalysisType.getText().equalsIgnoreCase(ConstantValues.resourceEfficiencyAnalysisText)) {
			HashMap<String, File> gapFiles;
			ArrayList<String> selectedGapTypes = new ArrayList<String>();
			//The type of gaps to be calculated will be depended on the final resource patterns that can be discovered using
			//event gap analysis approach
			selectedGapTypes.add(ConstantValues.activityGapType);
			
			gapFiles = gap.enrichGap(allWithinEvents, gapPath, context, 
						selectedGapTypes, true, "complete");
				//Need to make gapFiles as an array because prom plug-in cannot handle generic
			String[] fileNames = new String[gapFiles.size()];
			File[] files = new File[gapFiles.size()];
			int i=0;
			for (String fileName: gapFiles.keySet()) {
				fileNames[i] = fileName;
				files[i] = gapFiles.get(fileName);
				i++;
			}	
			context.log("DONE!");
			context.getProgress().inc();
			System.out.println("here here here 9");
			//GapViewer gv = new GapViewer(context, gapFiles);
			ArrayList<Object> data = new ArrayList<Object>();
			data.add(0,fileNames);
			data.add(1,files);
			GapViewer gv = new GapViewer(context, data);
			Controller controller = new GapController(data, gv);
			return gv;
		} else if (resourceAnalysisType.getText().equalsIgnoreCase(ConstantValues.resourceDecisionTreeText)) {
			System.out.println("Decision tree resource analysis selected");
			return null;
		} else */ 
		if (resourceAnalysisType.getText().equalsIgnoreCase(ConstantValues.resourceQueueAnalysisText)) {
			System.out.println("Queue analysis is selected");
						
			QueueViewer qv = new QueueViewer(context, log);
			QueueController controller = new QueueController(allWithinEvents, qv);
			return qv;
		}	
		return null;
	}

	private int calculateMaximumSteps( ArrayList<String> selectedGapTypes, ArrayList<Event> allWithinEvents) {
		//calculate the total number of mini steps needed
		//Start with just 1 step (the last manual increment towards the end of the
		//plug-in calculation
		int allSteps=1;
		int noOfGapType = selectedGapTypes.size();
		//Then add each step for each iteration for each gap to be calculated
		allSteps = allSteps + (noOfGapType * allWithinEvents.size());

		return allSteps;
	}

	
}

