package org.processmining.plugins.resourcePattern.queueAnalysis;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.TreeMap;

import org.apache.commons.math3.distribution.ExponentialDistribution;
import org.apache.commons.math3.distribution.LogisticDistribution;
import org.apache.commons.math3.distribution.NormalDistribution;
import org.apache.commons.math3.distribution.PoissonDistribution;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.plugins.resourcePattern.ActivityInstance;
import org.processmining.plugins.resourcePattern.ConstantValues;
import org.processmining.plugins.resourcePattern.utility.Event;

public class QueueAnalysis {
	private String scopeType;
	private String queueType;
	private String priorityAttributeName;
	private String startType;
	private String endType;
	private ArrayList<Event> events;
	private Integer workItemCounter;
	private String forgetFactorType;
	private String queueOneTreatment;
	private HashMap<String, TreeMap<Date, Integer>> queueLength;
	
	HashMap<String, ArrayList<Event>> scopeItems;
	
	public QueueAnalysis(ArrayList<Event> events, HashMap<String, String> actionParameters) {
		Collections.sort(events);
		this.events=events;
		this.scopeType = actionParameters.get(ConstantValues.chosenScopeTypeOption);
		this.queueType = actionParameters.get(ConstantValues.chosenQueueTypeOption);
		this.startType = actionParameters.get(ConstantValues.chosenBeginTypeOption);
		this.endType = actionParameters.get(ConstantValues.chosenEndTypeOption);
		this.priorityAttributeName = queueType.equalsIgnoreCase(ConstantValues.queueTypePriority)?
				actionParameters.get(ConstantValues.chosenPriorityAttributeNameOption):null;
		this.forgetFactorType = actionParameters.get(ConstantValues.chosenForgetFactor);
		this.queueOneTreatment = actionParameters.get(ConstantValues.chosenQueueOneTreatment);
		this.workItemCounter = 0; 
		this.scopeItems = buildScopeItems();	
		
	}
	
	public HashMap<String, TreeMap<Date, Integer>> getQueueLength() {
		return queueLength;
	}

	public void setQueueLength(HashMap<String, TreeMap<Date, Integer>> queueLength) {
		this.queueLength = queueLength;
	}

	private HashMap<String, ArrayList<Event>> buildScopeItems() {
		HashMap<String, ArrayList<Event>> scopeItems=null;
		if (scopeType.equalsIgnoreCase(ConstantValues.scopeTypeActivity)) {
			scopeItems = buildActivityScope();
		} else if (scopeType.equalsIgnoreCase(ConstantValues.scopeTypeWorkload)) {
			scopeItems = buildWorkloadScope();
		} else if (scopeType.equalsIgnoreCase(ConstantValues.scopeTypeCase)) {
			scopeItems = buildCaseScope();
		} 
		return scopeItems;
	}

	private HashMap<String, ArrayList<Event>> buildCaseScope() {
		HashMap<String, HashMap<String,ArrayList<Event>>> scopeItems = new HashMap<String, HashMap<String,ArrayList<Event>>>();
		HashMap<String, ArrayList<Event>> perCase = new HashMap<String, ArrayList<Event>>();
		scopeItems.put("caseScope", perCase);
		for (Event e: events) {
			workItemCounter++;
			e.setTaskInstance(workItemCounter);
			if (!scopeItems.get("caseScope").containsKey(e.getCaseID())) {
				ArrayList<Event> startEndList = new ArrayList<Event>();
				Event startEvent = e.deepCopy(e);
				Event endEvent = e.deepCopy(e);
				startEvent.setTransition("start");
				endEvent.setTransition("complete");
				//startEvent.setActivity("startEvent");
				//endEvent.setActivity("endEvent");
				startEndList.add(0,startEvent);
				startEndList.add(1,endEvent);
				
				scopeItems.get("caseScope").put(e.getCaseID(), startEndList);
				scopeItems.put("caseScope",perCase);
			} else {
				Date inTS = scopeItems.get("caseScope").get(e.getCaseID()).get(0).getTimestamp();
				Date outTS = scopeItems.get("caseScope").get(e.getCaseID()).get(1).getTimestamp();
				if (e.getTimestamp().before(inTS)) {
					scopeItems.get("caseScope").get(e.getCaseID()).get(0).setTimestamp(e.getTimestamp());
				} else if (e.getTimestamp().after(outTS)) {
					scopeItems.get("caseScope").get(e.getCaseID()).remove(1);
					Event laterEvent = e.deepCopy(e);
					laterEvent.setTransition("complete");
					scopeItems.get("caseScope").get(e.getCaseID()).add(1,laterEvent);
				} else {
					//do nothing just middle events
				}
			}
			assert scopeItems.get("caseScope").get(e.getCaseID()).size()==2;
		}
		
		HashMap<String,ArrayList<Event>> caseEvents = new HashMap<String, ArrayList<Event>>();
		ArrayList<Event> firstLastEventsInitial = new ArrayList<Event>();
		caseEvents.put("caseScope", firstLastEventsInitial);
		for (String caseID: scopeItems.get("caseScope").keySet()) {
			ArrayList<Event> firstLastEvents = scopeItems.get("caseScope").get(caseID);
			assert firstLastEvents.size()==2;
			System.out.println("First event: " + firstLastEvents.get(0).getCaseID() +","+firstLastEvents.get(0).getActivity() + "," + firstLastEvents.get(0).getTransition() + "," + firstLastEvents.get(0).getTimestamp() );
			System.out.println("Last event: " + firstLastEvents.get(1).getCaseID() +","+firstLastEvents.get(1).getActivity() + "," + firstLastEvents.get(1).getTransition() + "," + firstLastEvents.get(1).getTimestamp() );
			caseEvents.get("caseScope").addAll(firstLastEvents);
			System.out.println("Size is: " + caseEvents.get("caseScope").size());
		}
		return caseEvents;
	}

	private HashMap<String, ArrayList<Event>> buildWorkloadScope() {
		HashMap<String, ArrayList<Event>> scopeItems = new HashMap<String, ArrayList<Event>>();
		for (Event e: events) {
			if (!scopeItems.containsKey(e.getResource())) {
				ArrayList<Event> resourceItems = new ArrayList<Event>();
				scopeItems.put(e.getResource(), resourceItems);
			}
			if (e.getCaseID().equalsIgnoreCase("26825") && e.getActivity().equalsIgnoreCase("approvePO")) {
				System.out.println("building scope items found:" + e.getTransition());
			}
			workItemCounter++;
			e.setTaskInstance(workItemCounter);
			scopeItems.get(e.getResource()).add(e);
		}
		return scopeItems;
	}

	private HashMap<String, ArrayList<Event>> buildActivityScope() {
		HashMap<String, ArrayList<Event>> scopeItems = new HashMap<String, ArrayList<Event>>();
		for (Event e: events) {
			if (!scopeItems.containsKey(e.getActivity())) {
				ArrayList<Event> activityItems = new ArrayList<Event>();
				scopeItems.put(e.getActivity(), activityItems);
			}
			workItemCounter++;
			e.setTaskInstance(workItemCounter);
			scopeItems.get(e.getActivity()).add(e);
		}
		return scopeItems;
	}
	
	public HashMap<String, TreeMap<Integer, ActivityInstance>> assessQueue() throws Exception {
		HashMap<String, TreeMap<Integer, ActivityInstance>> queueScores = null;
		
		if (this.queueType.equalsIgnoreCase(ConstantValues.queueTypeFIFO)) {
			FIFOAnalysis fifo = new FIFOAnalysis(scopeItems, this.startType, this.endType, this.queueOneTreatment);
			queueScores = fifo.getQueuesScoreList();
			queueLength = fifo.getQueueLength();
			//queueScores = new FIFOAnalysis(scopeItems, this.startType, this.endType, this.queueOneTreatment).getQueuesScoreList();
		} else if (this.queueType.equalsIgnoreCase(ConstantValues.queueTypeLIFO)) {
			LIFOAnalysis lifo = new LIFOAnalysis(scopeItems, this.startType, this.endType, this.queueOneTreatment);
			queueScores = lifo.getQueuesScoreList();
			queueLength = lifo.getQueueLength();
			//queueScores = new LIFOAnalysis(scopeItems, this.startType, this.endType, this.queueOneTreatment).getQueuesScoreList();
		} else if (this.queueType.equalsIgnoreCase(ConstantValues.queueTypePriority)) {
			PriorityQueue priorityQ = new PriorityQueue(scopeItems, this.startType, this.endType, 
					this.priorityAttributeName, ConstantValues.queueTypePriorit_AlphabeticFIFO, this.queueOneTreatment);
			queueScores = priorityQ.getQueuesScoreList();
			queueLength = priorityQ.getQueueLength();
			//queueScores = new PriorityQueue(scopeItems, this.startType, this.endType, 
			//		this.priorityAttributeName, ConstantValues.queueTypePriorit_AlphabeticFIFO, this.queueOneTreatment).getQueuesScoreList();
		}
		if (queueScores!=null) {
			updateWithAdjustedScore(this.forgetFactorType, queueScores);
			
		} 
		
		for (String groupBykey: queueScores.keySet()) {
			TreeMap<Integer, ActivityInstance> perGroupScores = queueScores.get(groupBykey);
			for (Integer i: perGroupScores.keySet()) {
				ActivityInstance ai = perGroupScores.get(i);
				if (ai.getQueueScore()==null || ai.getAdjustedScore()==null) {
					System.out.println("serdfdas");
				}
				
			}
		}
		printToFile(queueScores);
		return queueScores;
	}

	private void printToFile(HashMap<String, TreeMap<Integer, ActivityInstance>> queueScores) throws IOException {
		File outFile = new File(ConstantValues.fileLocation+File.separator +"output.csv");
		if (outFile.exists()) {
			outFile.delete();
		}
		FileWriter fw = new FileWriter(outFile, true);
		BufferedWriter bw = new BufferedWriter(fw);
		bw.write("groupByKey,caseID,activity,resource,date,inIndex,outIndex,queueScore,adjustedScore\r\n");
		bw.flush();
		for (String groupByKey: queueScores.keySet()) {
			TreeMap<Integer,ActivityInstance> perGroupScores = queueScores.get(groupByKey);
			for (Integer i: perGroupScores.keySet()) {
				bw.write(groupByKey + "," + perGroupScores.get(i).toString()+"\r\n");
				bw.flush();

			}
		}
		
		bw.close();
		fw.close(); 
		
	}

	private void updateWithAdjustedScore(String chosenDistribution,
			HashMap<String, TreeMap<Integer, ActivityInstance>> queueScores) {
		
		for (String groupByKey: queueScores.keySet()) {
			TreeMap<Integer, ActivityInstance> perGroupScores = queueScores.get(groupByKey);
			DescriptiveStatistics ds = new DescriptiveStatistics();
			Date minOutTS = null;
			Date maxOutTS = null;
			for (ActivityInstance ai: perGroupScores.values()) {
				if (ai.getOutTS()==null) {
					System.out.println("No OUT TS");
					continue;
				}
				ds.addValue(ai.getOutTS().getTime());
				if (minOutTS==null) {
					minOutTS = ai.getOutTS();
				} else {
					minOutTS = minOutTS.before(ai.getOutTS())? minOutTS : ai.getOutTS(); 
				}
				
				if (maxOutTS == null) {
					maxOutTS = ai.getOutTS();
				} else {
					maxOutTS = maxOutTS.after(ai.getOutTS())? maxOutTS: ai.getOutTS();
				}
			}
			
			
			for (Integer i: perGroupScores.keySet()) {
				ActivityInstance ai = perGroupScores.get(i);
				
				//if we are ignoring queuelength of one, then we need to give the same marker with the 
				//adjusted score. Otherwise, either we use half mark (0.5) or normal calculation, the adjusted score will
				//flow naturally from the original score
				if (this.queueOneTreatment.equalsIgnoreCase(ConstantValues.ignoreCalculation) && ai.getQueueScore()==-1) {
					ai.setAdjustedScore((double) -1);
					continue;
				} 
				long currentSkip = ai.getOutTS().getTime();
				
				Double retainFactor_nonEndpoints=null;
				Double retainFactor=null;

				if (chosenDistribution.equalsIgnoreCase(ConstantValues.normalFF)) {
                          //NormalDistribution nd = new NormalDistribution(ds.getMean(), ds.getStandardDeviation());
                          NormalDistribution nd = new NormalDistribution(0, 1);
                          double normalizedValue = (( (double) currentSkip - (double) minOutTS.getTime()) / ( (double) maxOutTS.getTime() - (double) minOutTS.getTime()));
                          double normalizedSkip = (normalizedValue * 10) - 5;
                          //retainFactor_nonEndpoints = nd.cumulativeProbability(currentSkip);
                          retainFactor_nonEndpoints = nd.cumulativeProbability(normalizedSkip);
                   } else if (chosenDistribution.equalsIgnoreCase(ConstantValues.poissonFF)) {
                          //PoissonDistribution pd = new PoissonDistribution(ds.getMean(), 5);
                          PoissonDistribution pd = new PoissonDistribution(10);
                          double normalizedValue = (( (double) currentSkip - (double) minOutTS.getTime()) / ( (double) maxOutTS.getTime() - (double) minOutTS.getTime()));
                          double normalizedSkip = normalizedValue * 20;
                          retainFactor_nonEndpoints = pd.cumulativeProbability((int)normalizedSkip);
                   } else if (chosenDistribution.equalsIgnoreCase(ConstantValues.expFF)) {
                          //ExponentialDistribution ed = new ExponentialDistribution(ds.getMean());
                          ExponentialDistribution ed = new ExponentialDistribution(1);
                          double normalizedValue = (( (double) currentSkip - (double) minOutTS.getTime()) / ( (double) maxOutTS.getTime() - (double) minOutTS.getTime()));
                          double normalizedSkip = normalizedValue * 5;
                          //retainFactor_nonEndpoints = ed.cumulativeProbability(currentSkip);
                          retainFactor_nonEndpoints = ed.cumulativeProbability(normalizedSkip);
                   } else if (chosenDistribution.equalsIgnoreCase(ConstantValues.logisticExpFF)) {
                          //LogisticDistribution ld = new LogisticDistribution(null, ds.getMean(), 3);
                          LogisticDistribution ld = new LogisticDistribution(null, 10, 2);
                          double normalizedValue = (( (double) currentSkip - (double) minOutTS.getTime()) / ( (double) maxOutTS.getTime() - (double) minOutTS.getTime()));
                          double normalizedSkip = (normalizedValue * 25) - 5;
                          //retainFactor_nonEndpoints = ld.cumulativeProbability(currentSkip);
                          retainFactor_nonEndpoints = ld.cumulativeProbability(normalizedSkip);
                   } else if (chosenDistribution.equalsIgnoreCase(ConstantValues.numberEFF)) {
					double[] sortedValues = ds.getSortedValues();
					double position=0;
					int k=0;
					for (k=0; k<sortedValues.length; k++) {
						if (sortedValues[k]==currentSkip) {
							break;
						}
					}
					position = ds.getN() - k;
					System.out.println("Position is: " + position);
					retainFactor_nonEndpoints= (double) (1/(Math.pow(1.0001, position)));
				
				} else { 
					retainFactor_nonEndpoints = new Double(1);
				}
				if (retainFactor_nonEndpoints == null) {
					System.out.println("fdasfd");
				}
				
				if (ai.getOutTS().equals(minOutTS)) {
					retainFactor = new Double(0);
				//} else if (ai.getOutIndex().equals(maxOutTS)) {
				} else if (ai.getOutTS().equals(maxOutTS)) {
					retainFactor = new Double(1);
				} else {
					//double power = ((double)currentSkip/(double)maxTimespan) * (double) 100;
					//retainFactor = 1- (1/(Math.pow(forgetParameter, power)));
					//retainFactor = nd.cumulativeProbability(currentSkip);
					retainFactor = retainFactor_nonEndpoints;
					if (retainFactor == null) {
						System.out.println("fdasfd");
					}
					System.out.println("retain factor is: " + retainFactor);
				}
				
				ai.setRetainFactor(retainFactor);
				//ai.setAdjustedScore(ai.getQueueScore() * retainFactor);
				ai.setAdjustedScore(retainFactor * (1-ai.getQueueScore()));
				
				assert ai.getAdjustedScore()!=null;
				assert ai.getQueueScore()!=null;
			}

		}
	}
}
