package org.processmining.plugins.resourcePattern.queueAnalysis;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.processmining.plugins.resourcePattern.ActivityInstance;
import org.processmining.plugins.resourcePattern.ConstantValues;
import org.processmining.plugins.resourcePattern.WorkItem;
import org.processmining.plugins.resourcePattern.utility.Event;

public abstract class QueueAssessment {
	
	private String startType;
	private String endType;
	private String queueOneTreatment;
	HashMap<String, ArrayList<Event>> scopeItems;
	private HashMap<String, TreeMap<Date, Integer>> queueLength;

	
	protected HashMap<String, TreeMap<Date, Integer>> getQueueLength() {
		return queueLength;
	}

	protected void setQueueLength(HashMap<String, TreeMap<Date, Integer>> queueLength) {
		this.queueLength = queueLength;
	}

	public QueueAssessment(HashMap<String, ArrayList<Event>> scopeItems, String startType, String endType, String queueOneTreatment) {
		this.scopeItems = scopeItems;
		this.startType=startType;
		this.endType=endType;
		this.queueOneTreatment = queueOneTreatment;
		this.queueLength = new HashMap<String, TreeMap<Date, Integer>>();
	}
	
	public HashMap<String, TreeMap<Integer, ActivityInstance>> getQueuesScoreList() throws Exception{
		HashMap<String, TreeMap<Integer, ActivityInstance>> queuesScoreList = new HashMap<String, TreeMap<Integer, ActivityInstance>>();
		
		File calcFile = new File(ConstantValues.fileLocation + "/calculationFile.csv");
		if (calcFile.exists()) {
			calcFile.delete();
		}
		
		FileWriter fwc = new FileWriter(calcFile, true);
		BufferedWriter bwc = new BufferedWriter(fwc);
		
		File queueLengthFile = new File(ConstantValues.fileLocation + File.separator + "queuelength.csv");
		
	
		if (queueLengthFile.exists()) {
			queueLengthFile.delete();
		}
		
		
		
		FileWriter qfwc = new FileWriter(queueLengthFile, true);
		BufferedWriter qbwc = new BufferedWriter(qfwc);
		
		
		for(String groupByKey: scopeItems.keySet()) {
			ArrayList<Event> events = scopeItems.get(groupByKey);
			Collections.sort(events);
			
			//TreeMap<WorkItem, Integer> inQueue = new TreeMap<WorkItem, Integer>();
			//TreeMap<WorkItem, Integer> outQueue = new TreeMap<WorkItem, Integer>();
			//TreeMap<Integer, ActivityInstance> activityList = new TreeMap<Integer, ActivityInstance>();
			
			TreeMap<WorkItem, Integer> inQueue = new TreeMap<WorkItem, Integer>();
			TreeMap<WorkItem, Integer> outQueue = new TreeMap<WorkItem, Integer>();
			TreeMap<Integer, ActivityInstance> activityList = new TreeMap<Integer, ActivityInstance>();
			TreeMap<Date, Integer> queueLengthPerGroup = new TreeMap<Date, Integer>();
			
			int currentInPosition=0;
			int currentOutPosition=0;
			
			
			
			
			int index_E = 0;

			int queuelength = 0;
			
			int max_queuelength = 0;
			for (index_E = 0; index_E < events.size(); index_E ++)
			{
				Event e = events.get(index_E);
				
				if (e.getTransition().equalsIgnoreCase(this.startType))
				{
					queuelength ++;
					if (queuelength > max_queuelength)
					{
						max_queuelength = queuelength;
					}
				} else if (e.getTransition().equalsIgnoreCase(this.endType))
				{
					queuelength --;
				}
				
				//System.out.println("The queue length is " + queuelength+". The maximum queuelength is "+ max_queuelength+" transition type ="+e.getTransition());

			}
			
		//	System.out.println("Group by Key="+groupByKey + "; The maximum queuelength is "+ max_queuelength);
			qbwc.write(groupByKey+","+max_queuelength+"\r\n");
			qbwc.flush();
			
			
			
			
			for (Event e: events) {
				if (e.getCaseID().equalsIgnoreCase("26825") && e.getActivity().equalsIgnoreCase("approvePO")) {
					System.out.println("found:" + e.getTransition());
				}
			}
			
			for (Event e: events) {
				WorkItem wi = new WorkItem(e);
				if (e.getCaseID().equalsIgnoreCase("26825") && e.getActivity().equalsIgnoreCase("approvePO")) {
					System.out.println("stest osething here");
				}
				if (e.getTransition().equalsIgnoreCase(this.startType)) {
					//The if condition below is used to handle the existence of multiple concurrent
					//activity instances of the same type
					if (e.getCaseID().equalsIgnoreCase("26825") && e.getActivity().equalsIgnoreCase("approvePO")) {
						System.out.println("stest osething here");
					}
					if (inQueue.containsKey(wi)) {
						continue;
					}
					inQueue.put(wi, currentInPosition);
					ActivityInstance ai = new ActivityInstance(wi);
					ai.setInTS(e.getTimestamp());
					ai.setInIndex(currentInPosition);
					activityList.put(currentInPosition,ai);
					
					currentInPosition++;
				} else if (e.getTransition().equalsIgnoreCase(this.endType)) {
					if (e.getCaseID().equalsIgnoreCase("26825") && e.getActivity().equalsIgnoreCase("approvePO")) {
						System.out.println("stest osething here");
					}
					if (!inQueue.containsKey(wi)) {
						//we see an out work item without the corresponding in, incomplete log perhaps, just ignore
						continue;
					}
					outQueue.put(wi, currentOutPosition);
					Integer inWI = inQueue.get(wi);
					ActivityInstance aiTest = activityList.get(inQueue.get(wi));
					aiTest.setOutIndex(currentOutPosition);
					//activityList.get(inQueue.get(wi)).setOutIndex(currentOutPosition);
					activityList.get(inQueue.get(wi)).setOutTS(e.getTimestamp());
					
					Integer expectedOutputIndex = getExpectedOutputPosition(inQueue, activityList);

					Integer furthestIndex = getFurthestOutputPosition(inQueue, activityList);
					Integer closestIndex = getClosestOutputPosition(inQueue, activityList);
					
					Integer seenIndex = inQueue.get(wi);
					Integer distance = Math.abs(seenIndex-expectedOutputIndex);
					//Integer maxDistance = Math.abs(expectedOutputIndex - furthestIndex);
					Integer maxDistance = Math.abs(closestIndex-furthestIndex);
					
					//maxDistance will be zero if there is only one item in the inQueue
					//we are goign to ignore queue score if queue length is only one, so we put -1 as a flag for queue length 1
					//Double wiScore = maxDistance==0? -1 : ((double)distance/(double)maxDistance);
					Double wiScore = 0.0;
					if (maxDistance==0) {
//						if (this.queueOneTreatment == null)
//							System.out.println("queue one treatment is null");
						if (this.queueOneTreatment.equalsIgnoreCase(ConstantValues.normalCaluclation)) {
							wiScore = 0.0;
						} else if (this.queueOneTreatment.equalsIgnoreCase(ConstantValues.ignoreCalculation)) {
							wiScore = -1.0;
						} else if (this.queueOneTreatment.equalsIgnoreCase(ConstantValues.halfMarkCalculation)) {
							wiScore = 0.5;
						}
					} else {
						wiScore = ((double)distance/(double)maxDistance);
					}
					/*if (wiScore>1.0) {
						System.out.println("Something is not right");
					}*/	
					
					
					activityList.get(inQueue.get(wi)).setQueueScore(wiScore);
					activityList.get(inQueue.get(wi)).setCurrentInQueueAtOut(deepCopy(inQueue));
					
					//Store the queue length data per outTS
					queueLengthPerGroup.put(e.getTimestamp(), maxDistance+1);
					
					//Do clean up
					currentOutPosition++;;
					inQueue.remove(wi);
					
				}
			}
			
			//Now, need to transform the activity list queue, which is ordered based on the incoming queue, into one ordered
			//by the outgoing queue
			TreeMap<Integer, ActivityInstance> outScoreList = new TreeMap<Integer,ActivityInstance>();	
			
			for(Integer i: activityList.keySet()) {
				ActivityInstance ai = activityList.get(i);
				Integer index = ai.getOutIndex();
				if (index==null) {
					System.out.println("Something is not right, there is no corresponding 'out' ");
			//		incompleteAI.add(i);
					continue;
				}
				if (outScoreList.isEmpty() || (index!=null && !outScoreList.containsKey(index))) {
					outScoreList.put(index, activityList.get(index));
				} else {
					throw new Exception("Something is not right, there is out index duplicate");
				}
			}
			
			HashSet<Integer> incompleteAI = new HashSet<Integer>();
			for (Integer i: outScoreList.keySet()) {
				//bw.write(groupByKey+","+outScoreList.get(i).toString()+"\r\n");
				
			//	bwc.write(groupByKey+","+outScoreList.get(i).toString()+","+outScoreList.get(i).getOutTS()+"\n"+"The current queue was: \r\n");
				TreeMap<WorkItem, Integer> currentQueue = outScoreList.get(i).getCurrentInQueueAtOut();
				
				if (currentQueue==null) {
					System.out.println("Test");
					
					incompleteAI.add(i);
					continue;
				}
				
				TreeMap<Integer,WorkItem> flippedQueue = flipKeyValue(currentQueue);
				for (Integer index: flippedQueue.keySet()){
					WorkItem wi = flippedQueue.get(index);
					bwc.write(wi.getCaseID() + "," + wi.getActivity() + "," + wi.getResource() + "," + //outScoreList.get(i).getWi().getResource() + 
							//outScoreList.get(i).getInTS() + "," + outScoreList.get(i).getInIndex() + "," + outScoreList.get(i).getQueueScore() + "," + 
							//outScoreList.get(i).getAdjustedScore() + "\r\n");
							 activityList.get(currentQueue.get(wi)).getInTS()+ "," + activityList.get(currentQueue.get(wi)).getInIndex() + "," + 
							activityList.get(currentQueue.get(wi)).getOutTS()+ "," + 
							activityList.get(currentQueue.get(wi)).getQueueScore() + "," + activityList.get(currentQueue.get(wi)).getAdjustedScore() + "\r\n");
				}
				bwc.write(outScoreList.get(i).getQueueScore()+"\r\n\r\n");
			}
			//bw.flush();
			bwc.flush();
			for (Integer i: incompleteAI) {
				outScoreList.remove(i);
			}
			queuesScoreList.put(groupByKey, outScoreList);
			queueLength.put(groupByKey, queueLengthPerGroup);
			
		}
		//bw.close();
		//fw.close();
		bwc.close();
		fwc.close();
		
		
		qbwc.close();
		qfwc.close();
		
		return queuesScoreList;
	}
	
	public abstract Integer getClosestOutputPosition(TreeMap<WorkItem, Integer> inQueue,
			TreeMap<Integer, ActivityInstance> activityList);

	private TreeMap<WorkItem, Integer> deepCopy(TreeMap<WorkItem, Integer> inQueue) {
		TreeMap<WorkItem,Integer> newCopy = new TreeMap<WorkItem, Integer>();
		for (WorkItem wi: inQueue.keySet()) {
			//Event e = wi.getOriginalEvent();
			WorkItem newWI = wi.deepCopy(wi);
			newCopy.put(newWI, inQueue.get(wi));
		}
		return newCopy;
	}

	private TreeMap<Integer, WorkItem> flipKeyValue(TreeMap<WorkItem, Integer> currentQueue) {
		TreeMap<Integer,WorkItem> flipped = new TreeMap<Integer,WorkItem>();
		for (WorkItem wi: currentQueue.keySet()) {
			flipped.put(currentQueue.get(wi), wi);
		}
		return flipped;
	}

	public TreeSet<Integer> getValuesAsTreeSet(TreeMap<WorkItem,Integer> queue) {
		TreeSet<Integer> orderedSet = new TreeSet<Integer>();
		for(WorkItem wi: queue.keySet()) {
			orderedSet.add(queue.get(wi));
		}
		return orderedSet;
	}
	
	public abstract Integer getExpectedOutputPosition(TreeMap<WorkItem, Integer> inQueue, 
			TreeMap<Integer, ActivityInstance> outputScoreList);
	
	public abstract Integer getFurthestOutputPosition(TreeMap<WorkItem, Integer> inQueue, 
			TreeMap<Integer, ActivityInstance> outputScoreList);


}
