package org.processmining.plugins.resourcePattern.queueAnalysis.presentation.mvc;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeMap;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;

import org.deckfour.xes.model.XLog;
import org.processmining.plugins.resourcePattern.ActivityInstance;
import org.processmining.plugins.resourcePattern.ConstantValues;
import org.processmining.plugins.resourcePattern.utility.Helper;

public class QueueOptionPanel extends JPanel {

	private static final long serialVersionUID = 1221156771189559028L;
	private HashMap<String, TreeMap<Integer,ActivityInstance>> results;
	private final JLabel optionLabel = new JLabel("Options");

	private final JLabel queueTypeLabel = new JLabel("Queue Type:");
	private String[] queueType = new String[] {ConstantValues.queueTypeFIFO, ConstantValues.queueTypeLIFO, ConstantValues.queueTypePriority};
	private JComboBox<String> queueTypeComboBox = new JComboBox<String>(queueType);
	
	private final JLabel scopeLabel = new JLabel("Scope of queue:");
	private final String[] scopeType = { ConstantValues.scopeTypeActivity, ConstantValues.scopeTypeCase, ConstantValues.scopeTypeWorkload};
	private final JComboBox<String> scopeTypeComboBox = new JComboBox<String>(scopeType);

	private final JLabel beginTypeLabel = new JLabel("Begin transaction type:");
	private JComboBox<String> beginTypeBox=null;
	
	private final JLabel endTypeLabel = new JLabel("End transaction type:");
	private JComboBox<String> endTypeBox=null;
	
	private final JLabel forgetLabel = new JLabel("Type of forget factor");
	private JRadioButton noFF = new JRadioButton("No forget factor");
	private JRadioButton normalDistribitionFF = new JRadioButton("Normal Distribution CDF");
	private JRadioButton poissonDistributionFF = new JRadioButton("Poisson Distribution CDF");
	private JRadioButton exponentialDistributionFF = new JRadioButton("Exponential CDF");
	private JRadioButton logisticDistributionFF = new JRadioButton("Logistic CDF");
	private JRadioButton logNormalDistributionFF = new JRadioButton("Inverse Exponential (Normalized 0-1)");
	private ButtonGroup bg = new ButtonGroup();
	
	
	private final JLabel priorityAttributeLabel = new JLabel("Priority attribute name:");
	private JComboBox<String> priorityAttributeBox=null;
	
	private JLabel groupByChoiceLabel = new JLabel("Select the items to view:");
	private JList<String> groupByChoice = new JList<String>();

	private JLabel queueLengthOne = new JLabel("Calculation for queue of length 1:");
	private JRadioButton normalCalculationRadioButton = new JRadioButton("Normal");
	private JRadioButton ignoreCalculationRadioButton = new JRadioButton("Ignore");
	private JRadioButton halfMarkRadioButton = new JRadioButton("0.5 Score");
	private ButtonGroup queueOneButtonGroup = new ButtonGroup();
	
	private JCheckBox originalScoreCheck = new JCheckBox("Original Score Graph");
	private JCheckBox adjustedScoreCheck = new JCheckBox("Adjusted Score Graph");
	private JCheckBox movingAverageScoreCheck = new JCheckBox("Moving Average Graph");
	
	private final JButton getResultButton = new JButton("Display");
	private XLog log;

	private HashSet<String> resourceList=null;
	private HashSet<String> activityList = null;
	public QueueOptionPanel(XLog log) {
		super();
		this.log = log;
	}

	public JPanel getOptionPanel() {
		Helper helper = new Helper();
		HashSet<String> transactionTypeList = helper.getTransactionType(log);
		Set<String> attributeList = helper.getAvailableAttributes(log);
		this.resourceList = helper.getResourceList(log);
		this.activityList = helper.getActivityList(log);
		
		JPanel optionPanel = new JPanel();
		optionPanel.setBackground(Color.gray);
		optionPanel.setLayout(new BoxLayout(optionPanel, BoxLayout.Y_AXIS));
		optionPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.setPreferredSize(new Dimension(200, 300));

		//Set the panel title
		optionLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionLabel.setFont(new Font("SansSerif", Font.BOLD, 16));
		optionPanel.add(optionLabel);

		//Set the options
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		queueTypeLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(queueTypeLabel);
		queueTypeComboBox.setMaximumSize(new Dimension(160, 20));
		queueTypeComboBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		optionPanel.add(queueTypeComboBox);

		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		scopeLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(scopeLabel);
		scopeTypeComboBox.setMaximumSize(new Dimension(160, 20));
		scopeTypeComboBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		optionPanel.add(scopeTypeComboBox);

		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		beginTypeLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(beginTypeLabel);
		beginTypeBox = new JComboBox<String>(transactionTypeList.toArray(new String[transactionTypeList.size()]));
		beginTypeBox.setMaximumSize(new Dimension(160, 20));
		beginTypeBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		optionPanel.add(beginTypeBox);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		endTypeLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(endTypeLabel);
		endTypeBox = new JComboBox<String>(transactionTypeList.toArray(new String[transactionTypeList.size()]));
		endTypeBox.setMaximumSize(new Dimension(160, 20));
		endTypeBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		optionPanel.add(endTypeBox);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		endTypeLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(forgetLabel);
		optionPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		bg.add(noFF);
		bg.add(normalDistribitionFF);
		bg.add(exponentialDistributionFF);
		bg.add(poissonDistributionFF);
		bg.add(logisticDistributionFF);
		bg.add(logNormalDistributionFF);
		noFF.setSelected(true);
		optionPanel.add(noFF);
		optionPanel.add(normalDistribitionFF);
		optionPanel.add(exponentialDistributionFF);
		optionPanel.add(poissonDistributionFF);
		optionPanel.add(logisticDistributionFF);
		optionPanel.add(logNormalDistributionFF);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		priorityAttributeLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(priorityAttributeLabel);
		priorityAttributeBox = new JComboBox<String>(attributeList.toArray(new String[attributeList.size()]));
		priorityAttributeBox.setMaximumSize(new Dimension(160, 20));
		priorityAttributeBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		priorityAttributeBox.setEnabled(true);
		optionPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		optionPanel.add(priorityAttributeBox);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		queueOneButtonGroup.add(normalCalculationRadioButton);
		queueOneButtonGroup.add(ignoreCalculationRadioButton);
		queueOneButtonGroup.add(halfMarkRadioButton);
		normalCalculationRadioButton.setSelected(true);
		optionPanel.add(normalCalculationRadioButton);
		optionPanel.add(ignoreCalculationRadioButton);
		optionPanel.add(halfMarkRadioButton);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		groupByChoiceLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(groupByChoiceLabel);
		DefaultListModel<String> listmodel = new DefaultListModel<String>();
		listmodel.addElement("...");
		groupByChoice.setModel(listmodel);
		groupByChoice.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		groupByChoice.setLayoutOrientation(JList.VERTICAL);
		groupByChoice.setVisibleRowCount(3);
		groupByChoice.setEnabled(false);
		JScrollPane listScroller = new JScrollPane();
		listScroller.setViewportView(groupByChoice);
		//listScroller.setPreferredSize(new Dimension(150,50));
		listScroller.setEnabled(true);
		listScroller.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
		listScroller.setAlignmentX(Component.LEFT_ALIGNMENT);
		optionPanel.add(listScroller);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		optionPanel.add(originalScoreCheck);
		optionPanel.add(adjustedScoreCheck);
		optionPanel.add(movingAverageScoreCheck);
		
		optionPanel.add(Box.createRigidArea(new Dimension(5, 2)));
		optionPanel.add(getResultButton);

		return optionPanel;
	}	

	public JCheckBox getOriginalScoreCheck() {
		return originalScoreCheck;
	}

	public void setOriginalScoreCheck(JCheckBox originalScoreCheck) {
		this.originalScoreCheck = originalScoreCheck;
	}

	public JCheckBox getAdjustedScoreCheck() {
		return adjustedScoreCheck;
	}

	public void setAdjustedScoreCheck(JCheckBox adjustedScoreCheck) {
		this.adjustedScoreCheck = adjustedScoreCheck;
	}

	public JCheckBox getMovingAverageScoreCheck() {
		return movingAverageScoreCheck;
	}

	public void setMovingAverageScoreCheck(JCheckBox movingAverageScoreCheck) {
		this.movingAverageScoreCheck = movingAverageScoreCheck;
	}

	public JRadioButton getLogisticDistributionFF() {
		return logisticDistributionFF;
	}

	public void setLogisticDistributionFF(JRadioButton logisticDistributionFF) {
		this.logisticDistributionFF = logisticDistributionFF;
	}

	public JRadioButton getLogNormalDistributionFF() {
		return logNormalDistributionFF;
	}

	public void setLogNormalDistributionFF(JRadioButton logNormalDistributionFF) {
		this.logNormalDistributionFF = logNormalDistributionFF;
	}

	public JRadioButton getNormalCalculationRadioButton() {
		return normalCalculationRadioButton;
	}

	public void setNormalCalculationRadioButton(JRadioButton normalCalculationRadioButton) {
		this.normalCalculationRadioButton = normalCalculationRadioButton;
	}

	public JRadioButton getIgnoreCalculationRadioButton() {
		return ignoreCalculationRadioButton;
	}

	public void setIgnoreCalculationRadioButton(JRadioButton ignoreCalculationRadioButton) {
		this.ignoreCalculationRadioButton = ignoreCalculationRadioButton;
	}

	public JRadioButton getHalfMarkRadioButton() {
		return halfMarkRadioButton;
	}

	public void setHalfMarkRadioButton(JRadioButton halfMarkRadioButton) {
		this.halfMarkRadioButton = halfMarkRadioButton;
	}

	public JRadioButton getNoFF() {
		return noFF;
	}

	public void setNoFF(JRadioButton noFF) {
		this.noFF = noFF;
	}

	public JRadioButton getNormalDistribitionFF() {
		return normalDistribitionFF;
	}

	public void setNormalDistribitionFF(JRadioButton normalDistribitionFF) {
		this.normalDistribitionFF = normalDistribitionFF;
	}

	public JRadioButton getPoissonDistributionFF() {
		return poissonDistributionFF;
	}

	public void setPoissonDistributionFF(JRadioButton poissonDistributionFF) {
		this.poissonDistributionFF = poissonDistributionFF;
	}

	public JRadioButton getExponentialDistributionFF() {
		return exponentialDistributionFF;
	}

	public void setExponentialDistributionFF(JRadioButton exponentialDistributionFF) {
		this.exponentialDistributionFF = exponentialDistributionFF;
	}

	public ButtonGroup getBg() {
		return bg;
	}

	public void setBg(ButtonGroup bg) {
		this.bg = bg;
	}

	public JList<String> getGroupByChoice() {
		return groupByChoice;
	}

	public void setGroupByChoice(JList<String> groupByChoice) {
		this.groupByChoice = groupByChoice;
	}

	public HashSet<String> getResourceList() {
		return resourceList;
	}

	public void setResourceList(HashSet<String> resourceList) {
		this.resourceList = resourceList;
	}

	public HashSet<String> getActivityList() {
		return activityList;
	}

	public void setActivityList(HashSet<String> activityList) {
		this.activityList = activityList;
	}

	public JComboBox<String> getBeginTypeBox() {
		return beginTypeBox;
	}

	public JComboBox<String> getEndTypeBox() {
		return endTypeBox;
	}

	public JComboBox<String> getPriorityAttributeBox() {
		return priorityAttributeBox;
	}

	public void addSelectListener(ItemListener i) {
		this.queueTypeComboBox.addItemListener(i);
		this.scopeTypeComboBox.addItemListener(i);
	}

	public void addDisplayListener(ActionListener a) {
		this.getResultButton.addActionListener(a);
	}

	public JComboBox<String> getScopeTypeComboBox() {
		return scopeTypeComboBox;
	}

	public JComboBox<String> getQueueTypeComboBox() {
		return queueTypeComboBox;
	}

}
